




(* 
 * foldl : (’a * ’b -> ’b) -> ’b -> ’a list -> ’b
 * REQUIRES: f is total
 * ENSURES: foldl f b [x1, x2,..., xn] 
 *  =~ f (xn, f (..., f (x2, f (x1, b))...))
 *)
fun foldl (f:'a * 'b -> 'b) (b:'b) (lst:'a list) : 'b = 
  (case lst of
      [] => b 
    | x::xs => foldl f (f (x, b)) xs
  )

(* 
 * foldr : (’a * ’b -> ’b) -> ’b -> ’a list -> ’b
 * REQUIRES: f is total
 * ENSURES: foldr f b [x1, x2,..., xn] 
 *  =~ f (x1, f (x2, f (..., f (xn, b)...)))
 *)
fun foldr (f:'a * 'b -> 'b) (b:'b) (lst:'a list) : 'b = 
  (case lst of
      [] => b 
    | x::xs => f (x, foldr f b xs)
  )









(*
 * foldMap : ('a -> 'b) -> 'a list -> 'b list
 * REQUIRES: f is total
 * ENSURES: foldMap f [x1,x2,...,xn] =~ [f x1, f x2,..., f xn]
 *)
fun foldMap (f:'a -> 'b) : 'a list -> 'b list = 
  foldr (fn (x, ys) => (f x)::ys) []



(*
 * foldFilter : ('a -> bool) -> 'a list -> 'a list
 * REQUIRES: p is total
 * ENSURES: foldFilter p lst ==> a list of only 
 *  those elements of lst satisfying p
 *)
fun foldFilter (p:'a -> bool) = 
  foldr (fn (x, xs) => if p x then x::xs else xs) []
